#define TIMER_Q_C
/*
 *  Copyright (C) 1999, 2000, 2006 James Antill
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * email: james@and.org
 */
#include "main.h"

#include "mk.h"

MALLOC_CHECK_DECL();

unsigned int timer_q__walk_base_empty = 0;

static Timer_q_base *timer_q__start_base_norm = NULL;
static Timer_q_base *timer_q__start_base_empty = NULL;
static const struct timeval *timer_q__first_timeval = NULL;


/* ISO C magic, converts a ptr to ->next into ->next->prev */
#define TIMER_Q_CONV_PTR_NEXT_PREV(x) \
 ((Timer_q_node *)(((char *)(x)) - offsetof(Timer_q_node, next)))

#if !defined (NDEBUG) /* if debugging */

static void timer_q__chk_empties(void)
{
 Timer_q_base *scan = timer_q__start_base_empty;
 unsigned int num = 0;
 
  while (timer_q__walk_base_empty && scan)
  {
    if (scan->start || scan->quick_del)
      ++num;
    scan = scan->next;
  }

  ASSERT(num <= timer_q__walk_base_empty);
}
#define TIMER_Q__CHK_EMPTIES() timer_q__chk_empties();

static void timer_q__chk_base(Timer_q_base *base)
{
  Timer_q_node *scan = base->start;
  unsigned int num = 0;
  
  assert(!scan || !scan->has_prev || !((Timer_q_double_node *) scan)->prev);
  assert((base->start && base->end && !base->end->next) ||
         (!base->start && !base->end));
  
  while (scan)
  {
    Timer_q_node *scan_next = scan->next;
    
    assert(!scan_next || !scan->has_prev ||
           (((Timer_q_double_node *) scan_next)->prev == scan));

    ++num;
    scan = scan_next;
  }

  ASSERT(num == base->num);
}

# define TIMER_Q__CHECK(base) timer_q__chk_base(base)

static int inside_cb = FALSE;

# define TIMER_Q__INSIDE_CB_SET(x) inside_cb = (x)
# define TIMER_Q__INSIDE_CB_SAVE(x) int x = inside_cb; inside_cb = TRUE
# define TIMER_Q__INSIDE_CB_RESTORE(x) inside_cb = (x)

# define TIMER_Q__RESET_NODE(x) do { \
 if ((x)->has_prev) \
  if ((x)->has_func) { \
    Timer_q_double_func_node a = TIMER_Q_DEBUG_INIT_DOUBLE_FUNC_NODE; \
    *((Timer_q_double_func_node *) (x)) = a; } \
  else { \
    Timer_q_double_node a = TIMER_Q_DEBUG_INIT_DOUBLE_NODE; \
    *((Timer_q_double_node *) (x)) = a; } \
 else \
  if ((x)->has_func) { \
    Timer_q_func_node a = TIMER_Q_DEBUG_INIT_FUNC_NODE; \
    *((Timer_q_func_node *) (x)) = a; } \
  else { \
    Timer_q_node a = TIMER_Q_DEBUG_INIT_NODE; \
    *((Timer_q_node *) (x)) = a; } \
} while (FALSE)

#else /* if ! debugging */
# define TIMER_Q__CHK_EMPTIES() /* nothing */

# define TIMER_Q__CHECK(base) do { } while (FALSE)

# define TIMER_Q__INSIDE_CB_SET(x) do { } while (FALSE)
# define TIMER_Q__INSIDE_CB_SAVE(x) do { } while (FALSE)
# define TIMER_Q__INSIDE_CB_RESTORE(x) do { } while (FALSE)

# define TIMER_Q__RESET_NODE(x) do { } while (FALSE)
#endif

/* done when always safe to assign */
#define TQ__FIRST_CMP_LT(x, y)                  \
    (!(x) || (TIMER_Q_TIMEVAL_CMP(x, y) > 0))

#define TQ__MAYBE_FIRST_CMP_LT(x, y)            \
    ((x)  && (TIMER_Q_TIMEVAL_CMP(x, y) > 0))




/* forward reference... */
static Timer_q_base **timer_q__find_base(Timer_q_base *);


static void timer_q__add_node(Timer_q_base *base, Timer_q_node *passed,
                              void *data, const struct timeval *tv,
                              int flags, int do_free)
{
 int has_prev = FALSE;
 int has_func = FALSE;
 Timer_q_node **scan = NULL;
 
 TIMER_Q__CHECK(base);

 ++base->num;
 
 passed->tv = *tv;
 if (TQ__MAYBE_FIRST_CMP_LT(timer_q__first_timeval, tv))
   timer_q__first_timeval = &passed->tv;
 
 if (flags & TIMER_Q_FLAG_NODE_DOUBLE)
   has_prev = TRUE;
 if (flags & TIMER_Q_FLAG_NODE_FUNC)
 {
  if (flags & TIMER_Q_FLAG_NODE_DOUBLE)
    ((Timer_q_double_func_node *)passed)->func = base->func;
  else
    ((Timer_q_func_node *)passed)->func = base->func;
  has_func = TRUE;
 }
 
 passed->has_prev = has_prev;
 passed->has_func = has_func;
 passed->do_free = do_free;
 passed->quick_del = FALSE;
 passed->data = data;

 if (!base->start)
 {
  assert(!base->end);

  if (base->move_when_empty)
    ++timer_q__walk_base_empty;
  
  base->end = base->start = passed;
  passed->next = NULL;
  if (passed->has_prev)
    ((Timer_q_double_node *) passed)->prev = NULL;

  TIMER_Q__CHECK(base);
  return;
 }
 assert(base->end);

 /* always check the end, makes life easier (and quicker) */
 if (TIMER_Q_TIMEVAL_CMP(&passed->tv, &base->end->tv) >= 0)
 {
  if (passed->has_prev)
    ((Timer_q_double_node *) passed)->prev = base->end;

  base->end->next = passed;
  base->end = passed;
  passed->next = NULL;
  return;
 }

 if (TIMER_Q_TIMEVAL_CMP(&passed->tv, &base->start->tv) < 0)
 {
  if (base->start->has_prev)
    ((Timer_q_double_node *) base->start)->prev = passed;
  if (passed->has_prev)
    ((Timer_q_double_node *) passed)->prev = NULL;

  passed->next = base->start;
  base->start = passed;
  
  TIMER_Q__CHECK(base);
  return;
 }

 if (base->insert_from_end)
 {
   Timer_q_node *tmp = base->end;

   do
   {
     if (!tmp->has_prev)
       goto insert_from_beg;

     tmp = ((Timer_q_double_node *)tmp)->prev;
     assert(tmp);
   } while (TIMER_Q_TIMEVAL_CMP(&tmp->tv, &passed->tv) > 0);

  scan = &tmp->next;
 }
 else
 {
 insert_from_beg:
  scan = &base->start->next;

  assert(*scan);
  while (TIMER_Q_TIMEVAL_CMP(&(*scan)->tv, &passed->tv) < 0)
  {
   scan = &(*scan)->next;
   assert(*scan);
  }
 }
 
 if ((passed->next = *scan) && passed->next->has_prev)
   ((Timer_q_double_node *) passed->next)->prev = passed;
 
 *scan = passed;
 
 if (passed->has_prev)
   ((Timer_q_double_node *) passed)->prev = TIMER_Q_CONV_PTR_NEXT_PREV(scan);

 TIMER_Q__CHECK(base);
 
 return;
}

Timer_q_node *timer_q_add_node(Timer_q_base *base,
                               void *data, const struct timeval *tv,
                               int flags)
{
 Timer_q_node *node = NULL;

 assert(timer_q__find_base(base));
 assert(!base->quick_del);
 
 if (flags & (TIMER_Q_FLAG_NODE_SINGLE |
              TIMER_Q_FLAG_NODE_FUNC | TIMER_Q_FLAG_NODE_DOUBLE))
   /* nothing */ ;
 else
 {
  if (base->malloc_double)
    flags |= TIMER_Q_FLAG_NODE_DOUBLE;
  if (base->malloc_func)
    flags |= TIMER_Q_FLAG_NODE_FUNC;
 }
 
 if ((flags & (TIMER_Q_FLAG_NODE_DOUBLE | TIMER_Q_FLAG_NODE_FUNC)) == 
     (TIMER_Q_FLAG_NODE_DOUBLE | TIMER_Q_FLAG_NODE_FUNC))
   node = MK(sizeof(Timer_q_double_func_node));
 else if ((flags & (TIMER_Q_FLAG_NODE_DOUBLE | TIMER_Q_FLAG_NODE_FUNC)) == 
          (TIMER_Q_FLAG_NODE_DOUBLE))
   node = MK(sizeof(Timer_q_double_node));
 else if ((flags & (TIMER_Q_FLAG_NODE_DOUBLE | TIMER_Q_FLAG_NODE_FUNC)) == 
          (TIMER_Q_FLAG_NODE_FUNC))
   node = MK(sizeof(Timer_q_func_node));
 else
   node = MK(sizeof(Timer_q_node));
 
 if (!node)
   return (NULL);

 timer_q__add_node(base, node, data, tv, flags, TRUE);

 return (node);
}

Timer_q_node *timer_q_add_static_node(Timer_q_node *node, Timer_q_base *base,
                                      void *data, const struct timeval *tv,
                                      int flags)
{
 assert(timer_q__find_base(base));
 assert(!base->quick_del);
 
 if (node && (flags & (TIMER_Q_FLAG_NODE_SINGLE |
                       TIMER_Q_FLAG_NODE_DOUBLE |
                       TIMER_Q_FLAG_NODE_FUNC)))
 {
  assert(!node->next &&
         (node->tv.tv_sec == TIMER_Q_TYPE_CONST_DEBUG_32) &&
         (node->tv.tv_usec == TIMER_Q_TYPE_CONST_DEBUG_32));

  assert((flags & TIMER_Q_FLAG_NODE_DOUBLE) ? node->has_prev : !node->has_prev);
  assert((flags & TIMER_Q_FLAG_NODE_FUNC) ? node->has_func : !node->has_func);

  timer_q__add_node(base, node, data, tv, flags, FALSE);
  return (node);
 }

 assert(FALSE); /* make sure people can learn easily */
 
 return (timer_q_add_node(base, data, tv, flags));
}

static Timer_q_node **timer_q__find_data(Timer_q_base *base, void *data)
{
 Timer_q_node **scan = &base->start;
 
 assert(data);

 while (*scan)
 {
  if (data == (*scan)->data)
    return (scan);
  
  scan = &(*scan)->next;
 }

 return (NULL);
}

Timer_q_node *timer_q_find_data(Timer_q_base *scan, void *data)
{
 Timer_q_node **ret = NULL;

 if ((ret = timer_q__find_data(scan, data)))
   return (*ret);
 
 return (NULL);
}

static Timer_q_node **timer_q__find_node(Timer_q_base *base,
                                         Timer_q_node *node)
{
 Timer_q_node **scan = &base->start;

 while (*scan)
 {
  if (node == *scan)
    return (scan);
  
  scan = &(*scan)->next;
 }

 return (NULL);
}

static void timer_q__del_node(Timer_q_base *base, Timer_q_node **current,
                              int just_unlink)
{
 Timer_q_node *tmp = *current;

 if (timer_q__first_timeval == &tmp->tv)
   timer_q__first_timeval = NULL;
  
 if (base->start == *current)
 {
  if (base->end == *current)
  {
   assert(!(*current)->next);
   
   base->end = NULL;
  }
  else if ((*current)->next && (*current)->next->has_prev)
    ((Timer_q_double_node *) (*current)->next)->prev = NULL;
 }
 else
 {
  Timer_q_node *prev = NULL;
  
  prev = TIMER_Q_CONV_PTR_NEXT_PREV(current);

  if ((*current)->next && (*current)->next->has_prev)
  {
   assert(!(base->end == *current));

   ((Timer_q_double_node *) (*current)->next)->prev = prev;
  }
  else if (base->end == *current)
  {
   assert(!(*current)->next);

   base->end = prev;
  }
 }
 
 *current = (*current)->next;

 --base->num;
   
 if (!just_unlink)
 {
  void (*func)(int, void *);
  int do_free = tmp->do_free;
  void *data = tmp->data;

  if (tmp->has_func)
    if (tmp->has_prev)
      func = ((Timer_q_double_func_node *)tmp)->func;
    else
      func = ((Timer_q_func_node *)tmp)->func;
  else
    func = base->func;

  if (do_free)
    F(tmp);
  else
    TIMER_Q__RESET_NODE(tmp); /* so you can realloc inside the cb */
  
  (*func)(TIMER_Q_TYPE_CALL_DEL, data);
 }

 TIMER_Q__CHECK(base);
}

int timer_q_del_data(Timer_q_base *base, void *data)
{
 Timer_q_node **ret = NULL;
 
 ret = timer_q__find_data(base, data);

 if (ret)
   timer_q__del_node(base, ret, FALSE);

 return (ret != NULL);
}

void timer_q_del_node(Timer_q_base *base, Timer_q_node *current)
{
 Timer_q_node **scan = NULL;
 
 assert(timer_q__find_node(base, current));
 
 if (base->start == current)
   timer_q__del_node(base, &base->start, FALSE);
 else if (current->has_prev)
 { /* *start would == current otherwise */
  assert(((Timer_q_double_node *) current)->prev);
  timer_q__del_node(base,
                    &((Timer_q_double_node *) current)->prev->next, FALSE);
 }
 else
 {
  scan = timer_q__find_node(base, current);
    
  timer_q__del_node(base, scan, FALSE);
 }
}

void timer_q_quick_del_node(Timer_q_node *node)
{
  /* NOTE: don't reset timer_q__first_timeval because __gen_first_timeval
   * takes quick_del nodes into account */

  node->quick_del = TRUE;
}

static void timer_q__add_base(Timer_q_base *base,
                              void (*func)(int, void *), int flags, int do_free)
{
 base->func = func;
 
 base->num = 0;
 base->start = NULL;
 base->end = NULL;
 base->quick_del = FALSE;
 base->do_free = do_free;
 
 base->insert_from_end = !!(flags & TIMER_Q_FLAG_BASE_INSERT_FROM_END);
 base->run_all         = !!(flags & TIMER_Q_FLAG_BASE_RUN_ALL);
 base->malloc_double   = !!(flags & TIMER_Q_FLAG_BASE_MALLOC_DOUBLE);
 base->malloc_func     = !!(flags & TIMER_Q_FLAG_BASE_MALLOC_FUNC);

 if (flags & TIMER_Q_FLAG_BASE_MOVE_WHEN_EMPTY)
 {
  base->next = timer_q__start_base_empty;
  timer_q__start_base_empty = base;
  
  base->move_when_empty = TRUE;
 }
 else
 {
  base->next = timer_q__start_base_norm;
  timer_q__start_base_norm = base;
  
  base->move_when_empty = FALSE;
 }
}

Timer_q_base *timer_q_add_base(void (*func)(int, void *), int flags)
{
 Timer_q_base *base = MK(sizeof(Timer_q_base));
 
 if (!base)
   return (NULL);
 
 timer_q__add_base(base, func, flags, TRUE);

 return (base);
}

Timer_q_base *timer_q_add_static_base(Timer_q_base *base,
                                      void (*func)(int, void *), int flags)
{
 assert(base &&
        !base->next && (base->num == TIMER_Q_TYPE_CONST_DEBUG_16));
 
 timer_q__add_base(base, func, flags, FALSE);

 return (base);
}

static Timer_q_base **timer_q__find_base(Timer_q_base *base)
{
 Timer_q_base **scan = &timer_q__start_base_norm;

 /* can be a _quick_del base, as they might have upgraded to a _del */
 while (*scan)
 {
  if (base == *scan)
    return (scan);
  
  scan = &(*scan)->next;
 }

 scan = &timer_q__start_base_empty;

 while (*scan)
 {
  if (base == *scan)
    return (scan);
  
  scan = &(*scan)->next;
 }

 return (NULL);
}

static void timer_q__del_base(Timer_q_base **base)
{
 Timer_q_base *tmp = *base;
 
 *base = (*base)->next;

 while (tmp->start)
 {
   if (timer_q__first_timeval == &tmp->start->tv)
   {
     ASSERT(!inside_cb);
     timer_q__first_timeval = NULL;
   }
   
   timer_q__del_node(tmp, &tmp->start, FALSE);
 }
 
 if (tmp->do_free)
  F(tmp);
 else
 {
  Timer_q_base a = TIMER_Q_DEBUG_INIT_BASE;
  
  *tmp = a;
 }
}

void timer_q_del_base(Timer_q_base *base)
{
 Timer_q_base **scan = NULL;

 ASSERT(!inside_cb);

 scan = timer_q__find_base(base);

 assert(scan);
 
 timer_q__del_base(scan);
}

void timer_q_quick_del_base(Timer_q_base *base)
{
 /* -- assert(base->do_free); / * trust me, quick del on static objs is bad */
 
 assert(timer_q__find_base(base));
 assert(!base->quick_del);
 
 base->quick_del = TRUE;

 if (!base->start && base->move_when_empty)
   ++timer_q__walk_base_empty;

 /* NOTE: unlike quick_del_node() this isn't supposed to happen often,
    so gen_first_timeval skips them, and we ignore all the timers in here */
 timer_q__first_timeval = NULL;  
}

static Timer_q_base *timer_q__get_empties(void)
{
 Timer_q_base **scan = &timer_q__start_base_empty;

 TIMER_Q__CHK_EMPTIES();

 /* It's possible that timer_q__walk_base_empty > number of elements,
  * Eg. call SET_MOVE_WHEN_EMPTY with 0, then 1 repeatedly */
 while (timer_q__walk_base_empty && *scan)
 {
  if ((*scan)->start || (*scan)->quick_del)
  {
   Timer_q_base *tmp = *scan;

   *scan = tmp->next;

   tmp->next = timer_q__start_base_norm;
   timer_q__start_base_norm = tmp;

   --timer_q__walk_base_empty;
  }
  else  
    scan = &(*scan)->next;
 }
 
 timer_q__walk_base_empty = 0;

 return (timer_q__start_base_norm);
}

static unsigned int timer_q__run_norm(Timer_q_base *base,
                                      const struct timeval *tv)
{
 Timer_q_node *scan = base->start;
 int called = 0;
 
 TIMER_Q__CHECK(base);
 
 while ((scan = base->start) &&
        (scan->quick_del || (TIMER_Q_TIMEVAL_CMP(&scan->tv, tv) <= 0)))
 {
  assert(!scan->next || !scan->next->has_prev ||
         (((Timer_q_double_node *) scan->next)->prev == scan));
  assert(base->start == scan);
  
  if (scan->quick_del)
    timer_q__del_node(base, &base->start, FALSE);
  else
  {
   void (*func)(int, void *);
   int do_free = scan->do_free;
   void *data = scan->data;
   
   timer_q__del_node(base, &base->start, TRUE);
  
   ++called;
   
   if (scan->has_func)
     if (scan->has_prev)
       func = ((Timer_q_double_func_node *)scan)->func;
     else
       func = ((Timer_q_func_node *)scan)->func;
   else
     func = base->func;

   if (do_free)
     F(scan);
   else
     TIMER_Q__RESET_NODE(scan); /* so you can realloc inside the cb */

   (*func)(TIMER_Q_TYPE_CALL_RUN_NORM, data);
  }
 }

 TIMER_Q__CHECK(base);
 
 return (called);
}

unsigned int timer_q_run_norm(const struct timeval *tv)
{
 Timer_q_base **scan = NULL;
 unsigned int num = 0;
 
 ASSERT(timer_q_first_timeval() || TRUE);
 
 ASSERT(!inside_cb);
 
 if (TQ__MAYBE_FIRST_CMP_LT(timer_q__first_timeval, tv))
   return (0);
 
 TIMER_Q__INSIDE_CB_SET(TRUE);

 timer_q__get_empties();
 
 scan = &timer_q__start_base_norm;
 while (*scan)
 {
  if (!(*scan)->quick_del)
    num += timer_q__run_norm(*scan, tv);

  if ((*scan)->quick_del)
    timer_q__del_base(scan);
  else if (!(*scan)->start && (*scan)->move_when_empty)
  {
   Timer_q_base *tmp = *scan;

   *scan = tmp->next;

   tmp->next = timer_q__start_base_empty;
   timer_q__start_base_empty = tmp;
  }
  else  
    scan = &(*scan)->next;
 }

 TIMER_Q__INSIDE_CB_SET(FALSE);

 /* in case anything added for before the current run */
 if (timer_q_first_timeval())
   num += timer_q_run_norm(tv);
 
 return (num);
}

static unsigned int timer_q__run_all(Timer_q_base *base)
{ /* ignore timestamps */
 Timer_q_node *scan = base->start;
 int called = 0;
  
 TIMER_Q__CHECK(base);
  
 while (scan)
 {
  Timer_q_node *scan_next = scan->next;
  
  assert(!scan_next || !scan_next->has_prev ||
         (((Timer_q_double_node *) scan_next)->prev == scan));
  
  if (!scan->quick_del)
  {
   void (*func)(int, void *);
   
   ++called;
   
   if (scan->has_func)
     if (scan->has_prev)
       func = ((Timer_q_double_func_node *)scan)->func;
     else
       func = ((Timer_q_func_node *)scan)->func;
   else
     func = base->func;
   
   (*func)(TIMER_Q_TYPE_CALL_RUN_ALL, scan->data);
  }
  
  scan = scan_next;
 }

 return (called);
}

unsigned int timer_q_run_all(void)
{
 Timer_q_base *scan = timer_q__get_empties();
 int num = 0;
 
 TIMER_Q__INSIDE_CB_SAVE(prev_inside_cb);

 while (scan)
 {
  if (scan->run_all && !scan->quick_del)
    num += timer_q__run_all(scan);

  scan = scan->next;
 }

 TIMER_Q__INSIDE_CB_RESTORE(prev_inside_cb);
 
 return (num);
}

/* NOTE: this can return tv's for quick deleted nodes ... that's n/p.
 * as run_norm() will just delete them */
static const struct timeval *timer_q__gen_first_timeval(void)
{ /* can't use timer_q__get_empties() as that'll alter state */
  Timer_q_base *scan = timer_q__start_base_norm;
  unsigned int count = timer_q__walk_base_empty;
  const struct timeval *ret = NULL;
  
  TIMER_Q__CHK_EMPTIES();
  
  while (scan)
  {
    if (!scan->quick_del &&
        scan->start && TQ__FIRST_CMP_LT(ret, &scan->start->tv))
      ret = &scan->start->tv;
    
    scan = scan->next;
  }
  
  scan = timer_q__start_base_empty;
  while (scan && count)
  {
    if (scan->quick_del)
      --count;
    else if (scan->start && TQ__FIRST_CMP_LT(ret, &scan->start->tv))
      ret = &scan->start->tv;
    
    scan = scan->next;
  }
  
  return (ret);
}
#undef TQ__FIRST_CMP_LT

const struct timeval *timer_q_first_timeval(void)
{
  ASSERT(!inside_cb);
  
  ASSERT(!timer_q__first_timeval ||
         (timer_q__first_timeval == timer_q__gen_first_timeval()));
  
  if (!timer_q__first_timeval)
    timer_q__first_timeval = timer_q__gen_first_timeval();

  return (timer_q__first_timeval);
}


void timer_q_exit(void)
{
  Timer_q_base **scan = &timer_q__start_base_norm;
  
  timer_q__get_empties();

  ASSERT(!timer_q__start_base_empty);
  ASSERT(!inside_cb);
  
  TIMER_Q__INSIDE_CB_SET(TRUE);
  
  scan = &timer_q__start_base_norm;
  while (*scan)
  {
    ASSERT((*scan)->quick_del);
    
    if ((*scan)->quick_del)
      timer_q__del_base(scan);
  }
  
  TIMER_Q__INSIDE_CB_SET(FALSE);

  MALLOC_CHECK_EMPTY();
}
